
from textual.app import ComposeResult
from textual.widgets import Footer, Static, Log, TabbedContent, TabPane
from textual.screen import Screen
from textual.containers import Container, VerticalScroll
from textual.message import Message

from .iclient import localtimestring

from .vectorpn import VectorPane


class GroupTabPane(TabPane):

    class AddVector(Message):
        """pass new vector to the pane."""

        def __init__(self, vector):
            self.vector = vector
            super().__init__()

    def __init__(self, groupname, groupid):
        self.groupname = groupname
        super().__init__(groupname, id=groupid)

    def compose(self):
        "For every vector draw it"
        devicename = self.app.itemid.devicename
        device = self.app.indiclient[devicename]
        vectors = list(vector for vector in device.values() if vector.group == self.groupname and vector.enable)
        with VerticalScroll():
            for vector in vectors:
                yield VectorPane(vector)

    def on_group_tab_pane_add_vector(self, message: AddVector) -> None:
        "Add a vector to this tab"
        vector = message.vector
        # get the VerticalScroll containing the vectors
        vs = self.query_one(VerticalScroll)
        vs.mount(VectorPane(vector))



class GroupPane(Container):

    DEFAULT_CSS = """

        GroupPane {
            width: 100%;
            padding: 1;
            min-height: 10;
            }
        """


    class AddGroup(Message):
        """pass new group to the pane."""

        def __init__(self, groupname: str) -> None:
            self.groupname = groupname
            super().__init__()


    class DelVector(Message):
        "Delete this vector"

        def __init__(self, vector, vectorid):
            self.vector = vector
            self.vectorid = vectorid
            super().__init__()


    def compose(self):
        "Create the widget holding tabs of groups, each grouptab will contain its vectors"
        devicename = self.app.itemid.devicename
        device = self.app.indiclient[devicename]
        groupset = set(vector.group for vector in device.values() if vector.enable)
        grouplist = list(groupset)
        grouplist.sort()
        with TabbedContent(id="dev_groups"):
            for groupname in grouplist:
                groupid = self.app.itemid.set_group_id(groupname)
                yield GroupTabPane(groupname, groupid)


    def on_group_pane_add_group(self, message: AddGroup) -> None:
        groupname = message.groupname
        groupid = self.app.itemid.set_group_id(groupname)
        tc = self.query_one('#dev_groups')
        tc.add_pane(GroupTabPane(groupname, groupid))


    def on_group_pane_del_vector(self, message: DelVector) -> None:
        vector = message.vector
        vectorid = message.vectorid
        vectorwidget = self.query_one(f"#{vectorid}")
        vectorwidget.remove()
        # remove the vector id's
        self.app.itemid.clear_vector(vector)
        # vector removed, does its group need to be removed?
        groupset = set(v.group for v in vector.device.values() if v.enable)
        # get the group of the deleted vector
        grp = vector.group
        if grp not in groupset:
            # the grp no longer has enabled contents, and must be removed
            grpid = self.app.itemid.get_group_id(grp)
            if grpid is None:
                return
            tc = self.query_one("#dev_groups")
            tc.remove_pane(grpid)
            self.app.itemid.unset_group(vector.devicename, grp)


class MessagesPane(Container):

    DEFAULT_CSS = """

        MessagesPane {
            height: 6;
            background: $panel;
            border: mediumvioletred;
           }

        MessagePane > Log {
            width: 100%;
            height: 100%;
            background: $panel;
            scrollbar-background: $panel;
            scrollbar-corner-color: $panel;
            }

        """

    class ShowLogs(Message):
        """pass messages to the pane."""

        def __init__(self, messagelog: str) -> None:
            self.messagelog = messagelog
            super().__init__()

    def compose(self) -> ComposeResult:
        self.border_title = "Device Messages"
        yield Log()


    def on_mount(self):
        log = self.query_one("Log")
        devicename = self.app.itemid.devicename
        messages = self.app.indiclient[devicename].messages
        if messages:
            log.write_lines( reversed([ localtimestring(t) + "  " + m for t,m in messages]) )
        else:
            log.write(f"Messages from {devicename} will appear here")

    def on_messages_pane_show_logs(self, message: ShowLogs) -> None:
        log = self.query_one("Log")
        if log.line_count < 32:
            log.write_line(message.messagelog)
            return
        # if greater than 32, clear logs, and show the last eight
        # stored as a deque in indiclient[devicename]
        devicename = self.app.itemid.devicename
        log.clear()
        if not (self.app.indiclient is None):
            messages = list(self.app.indiclient[devicename].messages)
            mlist = reversed([ localtimestring(t) + "  " + m for t,m in messages ])
            log.write_lines(mlist)



class DeviceSc(Screen):
    """The class defining the device screen."""

    DEFAULT_CSS = """

        DeviceSc >#devicename {
           height: 1;
           background: $primary;
           color: $text;
           padding-left: 2;
           dock: top;
           }
        """

    ENABLE_COMMAND_PALETTE = False

    BINDINGS = [("m", "main", "Main Screen")]

    def __init__(self, devicename):
        "set devicename in connections module"
        self.app.itemid.devicename = devicename
        super().__init__()

    def compose(self) -> ComposeResult:
        devicename = self.app.itemid.devicename
        yield Static(devicename, id="devicename")
        yield Footer()
        yield MessagesPane(id="dev-messages-pane")
        yield GroupPane(id="dev-group-pane")


    def action_main(self) -> None:
        """Event handler called when m pressed."""
        self.app.indiclient.clientdata['devicesc'] = None
        self.app.itemid.devicename = None
        self.app.pop_screen()


    def action_show_tab(self, tab: str) -> None:
        """Switch to a new tab."""
        self.get_child_by_type(TabbedContent).active = tab





import sys, argparse, pathlib

from . import version
from .iterm import IPyTerm



def main():
    """The commandline entry point to run the terminal client."""

    parser = argparse.ArgumentParser(usage="indipyterm [options]",
                                     description="Terminal client to communicate to an INDI service.")
    parser.add_argument("--port", type=int, default=7624, help="Port of the INDI server (default 7624).")
    parser.add_argument("--host", default="localhost", help="Hostname/IP of the INDI server (default localhost).")
    parser.add_argument("--blobfolder", help="Optional folder where BLOB's will be saved.")
    parser.add_argument("--version", action="version", version=version)
    args = parser.parse_args()

    if args.blobfolder:
        try:
            blobfolder = pathlib.Path(args.blobfolder).expanduser().resolve()
        except Exception:
            print("Error: If given, the BLOB's folder should be an existing directory")
            return 1
        else:
            if not blobfolder.is_dir():
                print("Error: If given, the BLOB's folder should be an existing directory")
                return 1
    else:
        blobfolder = None

    # run the IPyTerm app
    app = IPyTerm(host=args.host, port=args.port, blobfolder=blobfolder)
    app.run()

    return 0


if __name__ == "__main__":
    sys.exit(main())





from textual.widgets import Static, Button, Switch
from textual.reactive import reactive
from textual.containers import Container, VerticalScroll
from textual.widget import Widget
from textual.message import Message

from .iclient import localtimestring

from .memberpn import SwitchMemberPane, TextMemberPane, LightMemberPane, NumberMemberPane, BlobMemberPane, NumberInputField, TextInputField




class VectorTime(Static):

    DEFAULT_CSS = """
        VectorTime {
            margin-left: 1;
            margin-right: 1;
            width: auto;
        }
        """

    vtime = reactive("")


    def watch_vtime(self, vtime):
        if vtime:
            self.update(vtime)


class VectorState(Static):

    DEFAULT_CSS = """
        VectorState {
            margin-right: 1;
            width: auto;
            }
        """

    vstate = reactive("")

    def watch_vstate(self, vstate):
        if vstate == "Ok":
            self.styles.background = "darkgreen"
            self.styles.color = "white"
        elif vstate == "Alert":
            self.styles.background = "red"
            self.styles.color = "white"
        elif vstate == "Busy":
            self.styles.background = "yellow"
            self.styles.color = "black"
        elif vstate == "Idle":
            self.styles.background = "black"
            self.styles.color = "white"
        else:
            return
        self.update(vstate)



class VectorTimeState(Widget):

    DEFAULT_CSS = """
        VectorTimeState {
            layout: horizontal;
            align: right top;
            height: 1;
            }

        VectorTimeState > Static {
            width: auto;
            }
        """

    vtime = reactive("")
    vstate = reactive("")

    def compose(self):
        "Draw the timestamp and state"
        yield Static("State:")
        yield VectorTime().data_bind(VectorTimeState.vtime)
        yield VectorState().data_bind(VectorTimeState.vstate)


class VectorMessage(Static):

    DEFAULT_CSS = """
        VectorMessage {
            margin-left: 1;
            margin-right: 1;
            height: 2;
            }
        """

    vmessage = reactive("")

    def watch_vmessage(self, vmessage):
        if vmessage:
            self.update(vmessage)


class VectorPane(Widget):

    DEFAULT_CSS = """
        VectorPane {
            layout: vertical;
            height: auto;
            background: $panel;
            border: mediumvioletred;
            }
        """

    vtime = reactive("")
    vstate = reactive("")
    vmessage = reactive("")


    class ShowTimestamp(Message):

        def __init__(self, timestamp: str) -> None:
            self.timestamp = timestamp
            super().__init__()


    class ShowState(Message):

        def __init__(self, state: str) -> None:
            self.state = state
            super().__init__()


    class ShowVmessage(Message):

        def __init__(self, vmessage: str) -> None:
            self.vmessage = vmessage
            super().__init__()

    class SubmitButtonmessage(Message):

        def __init__(self, sbmessage: str) -> None:
            self.sbmessage = sbmessage
            super().__init__()


    def __init__(self, vector):
        "This VectorPane has attribute self.vector and id of the vectorid"
        self.vector = vector
        vectorid = self.app.itemid.set_id(vector.name)
        super().__init__(id=vectorid)


    def compose(self):
        "Draw the vector"
        self.border_title = self.vector.label

        self.vtime = localtimestring(self.vector.timestamp)
        self.vstate = self.vector.state

        vts = VectorTimeState()
        vts.data_bind(VectorPane.vtime)
        vts.data_bind(VectorPane.vstate)

        yield vts

        # create vector message
        if self.vector.message:
            self.vmessage = localtimestring(self.vector.message_timestamp) + "  " + self.vector.message

        yield VectorMessage().data_bind(VectorPane.vmessage)

        if self.vector.vectortype == "SwitchVector":
            yield SwitchVector(self.vector)
        elif self.vector.vectortype == "LightVector":
            yield LightVector(self.vector)
        elif self.vector.vectortype == "TextVector":
            yield TextVector(self.vector)
        elif self.vector.vectortype == "NumberVector":
            yield NumberVector(self.vector)
        elif self.vector.vectortype == "BLOBVector":
            yield BLOBVector(self.vector)


    def on_vector_pane_show_timestamp(self, message: ShowTimestamp) -> None:
        self.vtime = message.timestamp

    def on_vector_pane_show_state(self, message: ShowState) -> None:
        self.vstate = message.state

    def on_vector_pane_show_vmessage(self, message: ShowVmessage) -> None:
        self.vmessage = message.vmessage

    def on_vector_pane_submit_buttonmessage(self, message: SubmitButtonmessage) -> None:
        "Get the submit button status message, and update it"
        if self.vector.vectortype == "BLOBVector":
            # BLOB vectors do not have a submit button, so set message
            # in vector message space
            self.vmessage = message.sbmessage
            return
        vectorid = self.app.itemid.get_id(self.vector.name)
        buttonstatus = self.query_one(f"#{vectorid}_submitmessage")
        buttonstatus.update(message.sbmessage)




class SwitchVector(Widget):

    DEFAULT_CSS = """
        SwitchVector {
            height: auto;
            }
        SwitchVector > .submitbutton {
            layout: horizontal;
            align: right middle;
            height: auto;
            }
        SwitchVector > .submitbutton > Button {
            margin-right: 1;
            width: auto;
            }
        SwitchVector > .submitbutton > Static {
            margin-right: 4;
            width: auto;
            }

        """

    def __init__(self, vector):
        self.vector = vector
        super().__init__()

    def compose(self):
        "Draw the switch vector members"
        members = self.vector.members()
        # draw a switch for each vector member
        for member in members.values():
            yield SwitchMemberPane(self.vector, member)

        # After the switches, for rw or wo vectors, create a submit button

        if self.vector.perm != "ro":
            with Container(classes="submitbutton"):
                # create a static string with submit button
                # the string will hold any buttonstatus message required on an update being
                # submitted and will have id vectorid_submitmessage
                yield Static("", id=f"{self.app.itemid.get_id(self.vector.name)}_submitmessage")
                yield Button("Submit")


    def on_switch_changed(self, event):
        """Enforce the rule, OneOfMany AtMostOne AnyOfMany"""
        if self.vector.perm == "ro":
            # ignore switch changes for read only vectors
            return
        # clear buttonstatus message
        buttonstatus = self.query_one(f"#{self.app.itemid.get_id(self.vector.name)}_submitmessage")
        buttonstatus.update("")
        if self.vector.rule == "AnyOfMany":
            # No need to enforce this
            return
        if not event.value:
            # switch turned off
            return
        switches = self.query(Switch)
        for s in switches:
            if s is event.switch:
                # s is the switch changed
                continue
            if s.value:
                # any switch other than the one changed must be off
                s.value = False


    async def on_button_pressed(self, event):
        "Get membername:value dictionary and send it to the server"
        if self.vector.perm == "ro":
            # No submission for read only vectors
            return
        buttonstatus = self.query_one(f"#{self.app.itemid.get_id(self.vector.name)}_submitmessage")
        switchpanes = self.query(SwitchMemberPane)
        memberdict = {}
        for sp in switchpanes:
            membername = sp.member.name
            switch = sp.query_one(Switch)
            if switch.value:
                memberdict[membername] = "On"
            else:
                memberdict[membername] = "Off"
        # Check at least one pressed if rule is OneOfMany
        if self.vector.rule == "OneOfMany":
            oncount = list(memberdict.values()).count("On")
            if oncount != 1:
                buttonstatus.update("Invalid, OneOfMany rule requires one On switch")
                return
        # Check no more than one pressed if rule is AtMostOne
        if self.vector.rule == "AtMostOne":
            oncount = list(memberdict.values()).count("On")
            if oncount > 1:
                buttonstatus.update("Invalid, AtMostOne rule allows only one On switch")
                return
        # send this to the server
        buttonstatus.update("")
        # set state to busy
        self.parent.vstate = "Busy"
        await self.vector.send_newSwitchVector(members=memberdict)



class TextVector(Widget):

    DEFAULT_CSS = """
        TextVector {
            height: auto;
            }

        TextVector > .submitbutton {
            layout: horizontal;
            align: right middle;
            height: auto;
            }
        TextVector > .submitbutton > Button {
            margin-right: 1;
            width: auto;
            }
        TextVector > .submitbutton > Static {
            margin-right: 4;
            width: auto;
            }

        """

    def __init__(self, vector):
        self.vector = vector
        super().__init__()

    def compose(self):
        "Draw the number vector members"
        members = self.vector.members()
        for member in members.values():
            yield TextMemberPane(self.vector, member)

        if self.vector.perm != "ro":
            with Container(classes="submitbutton"):
                yield Static("", id=f"{self.app.itemid.get_id(self.vector.name)}_submitmessage")
                yield Button("Submit")

    async def on_button_pressed(self, event):
        "Get membername:value dictionary"
        if self.vector.perm == "ro":
            # No submission for read only vectors
            return
        buttonstatus = self.query_one(f"#{self.app.itemid.get_id(self.vector.name)}_submitmessage")
        textpanes = self.query(TextMemberPane)
        memberdict = {}
        for tp in textpanes:
            membername = tp.member.name
            textfield = tp.query_one(TextInputField)
            if textfield.placeholder:
                continue
            memberdict[membername] = textfield.value
        # send this to the server
        buttonstatus.update("")
        # set state to busy
        self.parent.vstate = "Busy"
        await self.vector.send_newTextVector(members=memberdict)



class LightVector(Widget):

    DEFAULT_CSS = """
        LightVector {
            height: auto;
            }
        """

    def __init__(self, vector):
        self.vector = vector
        super().__init__()

    def compose(self):
        "Draw the light vector"
        members = self.vector.members()
        for member in members.values():
            yield LightMemberPane(self.vector, member)


class NumberVector(Widget):

    DEFAULT_CSS = """
        NumberVector {
            height: auto;
            }

        NumberVector > .submitbutton {
            layout: horizontal;
            align: right middle;
            height: auto;
            }
        NumberVector > .submitbutton > Button {
            margin-right: 1;
            width: auto;
            }
        NumberVector > .submitbutton > Static {
            margin-right: 4;
            width: auto;
            }

        """

    def __init__(self, vector):
        self.vector = vector
        super().__init__()

    def compose(self):
        "Draw the number vector members"
        members = self.vector.members()
        for member in members.values():
            yield NumberMemberPane(self.vector, member)

        if self.vector.perm != "ro":
            with Container(classes="submitbutton"):
                yield Static("", id=f"{self.app.itemid.get_id(self.vector.name)}_submitmessage")
                yield Button("Submit")

    async def on_button_pressed(self, event):
        "Get membername:value dictionary"
        if self.vector.perm == "ro":
            # No submission for read only vectors
            return
        buttonstatus = self.query_one(f"#{self.app.itemid.get_id(self.vector.name)}_submitmessage")
        numberpanes = self.query(NumberMemberPane)
        memberdict = {}
        for np in numberpanes:
            membername = np.member.name
            numberfield = np.query_one(NumberInputField)
            if not numberfield.value:
                continue
            memberdict[membername] = numberfield.value
        # send this to the server
        buttonstatus.update("")
        # set state to busy
        self.parent.vstate = "Busy"
        await self.vector.send_newNumberVector(members=memberdict)



class BLOBVector(Widget):

    DEFAULT_CSS = """
        BLOBVector {
            height: auto;
            }
        """

    def __init__(self, vector):
        self.vector = vector
        super().__init__()

    def compose(self):
        "Draw the BLOB vector"
        members = self.vector.members()
        for member in members.values():
            yield BlobMemberPane(self.vector, member)




import pathlib

from textual.app import App
from textual import on
from textual.app import ComposeResult
from textual.widgets import Footer, Static, Button, Log, Input
from textual.screen import Screen
from textual.containers import Container, HorizontalScroll, VerticalScroll, Center
from textual.message import Message

from .iclient import ItemID, IClient, localtimestring

from .devicesc import DeviceSc

from . import version



class DevicePane(Container):

    DEFAULT_CSS = """

            DevicePane {
                width: 100%;
                row-span: 2;
                background: $panel;
                border: dodgerblue;
                }

            DevicePane > Static {
                background: $boost;
                color: auto;
                margin-bottom: 1;
                padding: 1;
                }

             DevicePane > Button {
                width: 100%;
                }
        """

    class NewButton(Message):
        """Add a new button."""

        def __init__(self, devicename: str) -> None:
            self.devicename = devicename
            super().__init__()

    class ClearDevices(Message):
        pass

    class DelButton(Message):
        """Delete a button."""

        def __init__(self, deviceid: str) -> None:
            self.deviceid = deviceid
            super().__init__()

    def compose(self):
        self.border_title = "Devices"
        if self.app.indiclient is None:
            devices = 0
        else:
            devices = self.app.indiclient.enabledlen()
        # The number of enabled devices
        if not devices:
            yield Static("No Devices found", id="no-devices")
        else:
            for devicename in self.app.indiclient:
                deviceid = self.app.itemid.set_devicid(devicename)
                yield Button(devicename, variant="primary", classes="devices", id=deviceid)


    def on_device_pane_new_button(self, message: NewButton) -> None:
        devicename = message.devicename
        deviceid = self.app.itemid.set_devicid(devicename)
        self.remove_children("#no-devices")
        self.mount(Button(devicename, variant="primary", classes="devices", id=deviceid))

    def on_device_pane_clear_devices(self, message: ClearDevices) -> None:
        if self.query(".devices"):
            self.remove_children(".devices")
            self.mount(Static("No Devices found", id="no-devices"))

    def on_device_pane_del_button(self, message: DelButton) -> None:
        deviceid = message.deviceid
        self.remove_children(f"#{deviceid}")


    @on(Button.Pressed, ".devices")
    def choose_device(self, event):
        "Choose device from the button pressed"
        iclient = self.app.indiclient
        if iclient is None:
            return
        devicename = self.app.itemid.get_devicename(event.button.id)
        if not devicename:
            return
        if devicename not in iclient:
            # An unknown device
            return
        if not iclient[devicename].enable:
            # This device is disabled
            return
        # create a device screen, and store a reference to it
        # in the indiclient 'cliendata' dictionary
        iclient.clientdata['devicesc'] = DeviceSc(devicename)
        # push the devicesc to the top of the stack
        self.app.push_screen(iclient.clientdata['devicesc'])


class BlobPane(HorizontalScroll):

    DEFAULT_CSS = """

        BlobPane {
            background: $panel;
            border: greenyellow;
            }
        """

    def compose(self):
        self.border_title = "Set BLOB folder"
        yield BlobInput(placeholder="Set a Folder to receive BLOBs", id="blob-input")

    def on_mount(self):
        iclient = self.app.indiclient
        if iclient is None:
            self.border_subtitle = "Received BLOBs disabled"
            return
        if iclient.BLOBfolder:
            self.border_subtitle = "Received BLOBs enabled"
        else:
            self.border_subtitle = "Received BLOBs disabled"


class BlobInput(Input):

    def on_mount(self):
        if self.app.blobfolder:
            self.insert_text_at_cursor(str(self.app.blobfolder))

    def on_blur(self, event):
        iclient = self.app.indiclient
        if not self.value:
            self.app.blobfolder = None
            if not (iclient is None):
                iclient.BLOBfolder = None
            self.clear()
            self.insert_text_at_cursor('')
            self.parent.border_subtitle = "Received BLOBs disabled"
            return

        blobfolder = pathlib.Path(self.value).expanduser().resolve()
        if not blobfolder.is_dir():
            self.app.blobfolder = None
            if not (iclient is None):
                iclient.BLOBfolder = None
            self.clear()
            self.insert_text_at_cursor('Invalid Folder')
            self.parent.border_subtitle = "Received BLOBs disabled"
            return

        self.app.blobfolder = blobfolder
        if not (iclient is None):
            iclient.BLOBfolder = blobfolder
        self.clear()
        self.insert_text_at_cursor(str(blobfolder))
        self.parent.border_subtitle = "Received BLOBs enabled"


    def action_submit(self):
        self.screen.focus_next('*')



class MessagesPane(Container):

    DEFAULT_CSS = """

        MessagesPane {
            width: 100%;
            column-span: 2;
            background: $panel;
            border: mediumvioletred;
            }

        MessagesPane > Log {
            width: 100%;
            background: $panel;
            scrollbar-background: $panel;
            scrollbar-corner-color: $panel;
            }
        """

    class ShowLogs(Message):
        """pass messages to the pane."""

        def __init__(self, messagelog: str) -> None:
            self.messagelog = messagelog
            super().__init__()

    def compose(self):
        self.border_title = "System Messages"
        yield Log(id="system-messages")

    def on_messages_pane_show_logs(self, message: ShowLogs) -> None:
        log = self.query_one("#system-messages")
        if log.line_count < 32:
            log.write_line(message.messagelog)
            return
        # if greater than 32, clear logs, and show the last eight
        # stored as a deque in indiclient
        log.clear()
        if not (self.app.indiclient is None):
            messages = list(self.app.indiclient.messages)
            mlist = reversed([ localtimestring(t) + "  " + m for t,m in messages ])
            log.write_lines(mlist)



class ConnectionPane(Container):

    DEFAULT_CSS = """

        ConnectionPane {
            background: $panel;
            border: mediumvioletred;
            align: center middle;
            }

        ConnectionPane > Static {
            margin: 1;
            }
        """


    def compose(self):
        self.border_title = "Set INDI Server"
        con_input = ConInput(placeholder="Host:Port", id="con-input")
        con_status = Static("Host:Port not set", id="con-status")
        con_button = Button("Connect", id="con-button")
        if self.app.indiclient is None:
            # No indiclient exists, enable the input field to accept a host and port
            con_input.disabled = False
            con_button.label = "Connect"
            if (not self.app.indihost) or (not self.app.indiport):
                con_status.update("Host:Port not set")
                con_button.disabled = True
            else:
                con_status.update(f"Current server : {self.app.indihost}:{self.app.indiport}")
                con_button.disabled = False
        else:
            # An indiclient instance exists, disable the input field
            # and set the button to 'Disconnect"
            con_input.disabled = True
            con_status.update(f"Current server : {self.app.indiclient.indihost}:{self.app.indiclient.indiport}")
            con_button.label = "Disconnect"
            con_button.disabled = False
        yield con_input
        yield con_status
        with Center():
            yield con_button


    async def on_button_pressed(self, event):
        con_input = self.query_one("#con-input")
        con_status = self.query_one("#con-status")
        con_button = self.query_one("#con-button")
        if self.app.indiclient is None:
            # call for connection
            # create an indiclient
            self.app.indiclient = IClient(indihost=self.app.indihost, indiport=self.app.indiport, app=self.app)
            if self.app.blobfolder:
                self.app.indiclient.BLOBfolder = self.app.blobfolder
            con_input.disabled = True
            con_status.update(f"Current server : {self.app.indiclient.indihost}:{self.app.indiclient.indiport}")
            con_button.label = "Disconnect"
            con_button.disabled = False
            # clear the messages pane
            mess_pane = self.parent.query_one("#sys-messages-pane")
            log = mess_pane.query_one("#system-messages")
            log.clear()
            # and run indiclient.asyncrun()
            self.app.run_worker(self.app.indiclient.asyncrun(), exclusive=True)
        else:
            # call for disconnection
            self.app.indiclient.shutdown()
            # and wait for it to shutdown
            await self.app.indiclient.stopped.wait()
            self.app.indiclient = None
            con_input.disabled = False
            con_status.update("Host:Port not set")
            con_button.label = "Connect"
            if (not self.app.indihost) or (not self.app.indiport):
                con_status.update("Host:Port not set")
                con_button.disabled = True
            else:
                con_status.update(f"Current server : {self.app.indihost}:{self.app.indiport}")
                con_button.disabled = False
            # clear the list of device buttons
            device_pane = self.parent.query_one("#device-pane")
            if device_pane.query(".devices"):
                device_pane.remove_children(".devices")
                device_pane.mount(Static("No Devices found", id="no-devices"))
            # clear the messages pane, leving a single 'DISCONNECTED' message
            mess_pane = self.parent.query_one("#sys-messages-pane")
            log = mess_pane.query_one("#system-messages")
            log.clear()
            log.write_line("DISCONNECTED")
            # and clear all item id's
            self.app.itemid.clear()



class ConInput(Input):

    def on_blur(self, event):

        hostport = self.value.strip()
        if hostport:
            hostportlist = hostport.split(":")
            if len(hostportlist) == 2:
                host = hostportlist[0].strip()
                port = hostportlist[1].strip()
                if not host:
                    host = "localhost"
                if not port.isdigit():
                    port = "7624"
                hostport = f"{host}:{port}"
            else:
                host = hostportlist[0].strip()
                if host:
                    hostport = host +":7624"
                else:
                    hostport = "localhost:7624"
        else:
            hostport = "localhost:7624"
        self.clear()
        self.insert_text_at_cursor(hostport)
        # set this new host and port into self.app.indihost, self.app.indiport
        self.app.indihost, self.app.indiport = hostport.split(":")
        # and enable the connection button
        con_button = self.parent.query_one("#con-button")
        con_button.disabled = False

    def action_submit(self):
        self.screen.focus_next('*')





class StartSc(Screen):
    """The top start screen."""

    DEFAULT_CSS = """

        StartSc > #title {
           background: $primary;
           color: $text;
           padding-left: 2;
           dock: top;
           }

        StartSc > #startsc-grid {
            height: 100%;
            min-height: 24;
            layout: grid;
            grid-size: 2 3;  /* two columns, three rows */
            grid-columns: 1fr 2fr;
            grid-rows: 2fr 1fr 1fr;
            }
        """

    ENABLE_COMMAND_PALETTE = False

    def __init__(self):
        super().__init__()


    def compose(self) -> ComposeResult:
        yield Static(f"INDI Terminal {version}", id="title")
        yield Footer()
        with VerticalScroll(id="startsc-grid"):
            yield DevicePane(id="device-pane")
            yield ConnectionPane(id="con-pane")
            yield BlobPane(id="blob-pane")
            yield MessagesPane(id="sys-messages-pane")


class IPyTerm(App):
    """An INDI terminal."""

    SCREENS = {"startsc": StartSc}

    BINDINGS = [("q", "quit", "Quit"), ("d", "toggle_dark", "Toggle dark mode")]

    ENABLE_COMMAND_PALETTE = False

    def __init__(self, host="localhost", port=7624, blobfolder=None):
        self.indihost = host
        self.indiport = port
        if blobfolder:
            bf = pathlib.Path(blobfolder).expanduser().resolve()
            if bf.is_dir():
                self.blobfolder = bf
            else:
                self.blobfolder = None
        else:
            self.blobfolder = None
        self.itemid = ItemID()
        self.indiclient = IClient(indihost=host, indiport=port, app=self)
        if self.blobfolder:
            self.indiclient.BLOBfolder = self.blobfolder
        super().__init__()


    def on_mount(self) -> None:
        """Start the worker which runs self.indiclient.asyncrun()
           and show the start screen"""
        self.push_screen('startsc')
        self.run_worker(self.indiclient.asyncrun(), exclusive=True)



    async def action_quit(self) -> None:
        """An action to quit the program."""
        if not (self.indiclient is None):
            self.indiclient.shutdown()
            # and wait for it to shutdown
            await self.indiclient.stopped.wait()
        self.exit(0)

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.theme = (
            "textual-dark" if self.theme == "textual-light" else "textual-light"
            )




import asyncio, logging

from datetime import datetime, timezone

import indipyclient as ipc


logger = logging.getLogger()
logger.addHandler(logging.NullHandler())



def localtimestring(t):
    "Return a string of the local time (not date)"
    localtime = t.astimezone(tz=None)
    # convert microsecond to integer between 0 and 100
    ms = localtime.microsecond//10000
    return f"{localtime.strftime('%H:%M:%S')}.{ms:0>2d}"



class ItemID():

    def __init__(self):
        self._itemdict = {}
        self._groupdict = {}
        # Every device, vector, widget will be given an id
        # starting with characters 'id' followed by a string number
        # created by incrementing this self._itemid
        self._itemid = 0

        # devicename needs to be set
        self.devicename = None

    def __bool__(self):
        return bool(self._itemdict)

    def get_group_id(self, groupname):
        if self.devicename is None:
            return
        if not groupname:
            raise KeyError("A group name must be given to get a group id")
        idnumber = self._groupdict.get((self.devicename, groupname))
        if idnumber is None:
            return
        return "gid"+str(idnumber)


    def set_group_id(self, groupname):
        if self.devicename is None:
            return
        if not groupname:
            raise KeyError("A group name must be given to set a group id")
        idnumber = self._groupdict.get((self.devicename, groupname))
        if idnumber is None:
            self._itemid += 1
            self._groupdict[self.devicename, groupname] = self._itemid
            return "gid"+str(self._itemid)
        return "gid"+str(idnumber)


    def unset_group(self, devicename, groupname):
        if not devicename:
            raise KeyError("A devicename must be given to unset a group id")
        if not groupname:
            raise KeyError("A group name must be given to unset a group id")
        self._groupdict[devicename, groupname] = None


    def get_id(self, vectorname=None, membername=None):
        if self.devicename is None:
            return
        if not vectorname:
            vectorname = None
        if not membername:
            membername = None
        if membername and (not vectorname):
            raise KeyError("If a membername is specified, a vectorname must also be given")
        idnumber = self._itemdict.get((self.devicename, vectorname, membername))
        if idnumber is None:
            return
        return "id"+str(idnumber)


    def set_id(self, vectorname=None, membername=None):
        "This create ids for widgets, and returns the id"
        if self.devicename is None:
            return
        if not vectorname:
            vectorname = None
        if not membername:
            membername = None
        if membername and (not vectorname):
            raise KeyError("If a membername is specified, a vectorname must also be given")
        idnumber = self._itemdict.get((self.devicename, vectorname, membername))
        if idnumber is None:
            self._itemid += 1
            self._itemdict[self.devicename, vectorname, membername] = self._itemid
            return "id"+str(self._itemid)
        return "id"+str(idnumber)


    def unset(self, devicename, vectorname=None, membername=None):
        if not vectorname:
            vectorname = None
        if not membername:
            membername = None
        if not devicename:
            raise KeyError("A devicename must be given to unset an id")
        if membername and (not vectorname):
            raise KeyError("If a membername is specified, a vectorname must also be given")
        self._itemdict[devicename, vectorname, membername] = None


    def get_devicid(self, devicename):
        if devicename is None:
            return
        idnumber = self._itemdict.get((devicename, None, None))
        if idnumber is None:
            return
        return "id"+str(idnumber)


    def set_devicid(self, devicename):
        "This create id for a device"
        if devicename is None:
            return
        idnumber = self._itemdict.get((devicename, None, None))
        if idnumber is None:
            self._itemid += 1
            self._itemdict[devicename, None, None] = self._itemid
            return "id"+str(self._itemid)
        return "id"+str(idnumber)


    def clear_device(self, device):
        "clear the id's of device, and its vectors and members"
        self.unset(device.devicename)
        for vectorname in device:
            self.unset(device.devicename, vectorname)
            self.unset_group(device.devicename, device[vectorname].group)
            membernamelist = list(device[vectorname].keys())
            for membername in membernamelist:
                self.unset(device.devicename, vectorname, membername)


    def clear_vector(self, vector):
        "delete the ids of the vector and all its members"
        self.unset(vector.devicename, vector.name)
        membernamelist = list(vector.keys())
        for membername in membernamelist:
            self.unset(vector.devicename, vector.name, membername)


    def get_devicename(self, deviceid):
        "Given an id, get the devicename, or return None if it does not exist"
        idnumber = int(deviceid.strip("id"))
        for key,value in self._itemdict.items():
            if value == idnumber:
                return key[0]

    def clear(self):
        self._itemdict.clear()
        self._groupdict.clear()
        self._itemid = 0



class IClient(ipc.IPyClient):

    async def rxevent(self, event):
        app = self.clientdata['app']
        startsc = app.get_screen('startsc')

        # if the connection is failed
        # ensure the startsc is being shown
        # and all device and vector id's are cleared

        if (not self.connected) and app.itemid :
            # the connection is disconnected
            if not startsc.is_active:
                app.push_screen('startsc')
            device_pane = startsc.query_one("#device-pane")
            device_pane.post_message(device_pane.ClearDevices())
            app.itemid.clear()
            self.clientdata['devicesc'] = None

        # handle received events affecting startsc ################################

        if (event.eventtype == "Define" or event.eventtype == "DefineBLOB"):
            # does this device have an id
            devicename = event.devicename
            if not app.itemid.get_devicid(devicename):
                # it doesn't, so add a button to the devicepane of startsc
                device_pane = startsc.query_one("#device-pane")
                device_pane.post_message(device_pane.NewButton(devicename))
                # As this is a new device, its devicesc cannot be currently showing,
                # but a message can be added to the device, which will appear on devicesc
                event.device.messages.appendleft( (event.timestamp, f"Device discovered: {devicename}") )
                return

            # this devicename already exists, but the Define event will be
            # adding a vector so that is still to do

        if (event.eventtype == "Message") and (not event.devicename):
            # This is a system message and should be added to the messages pane of startsc
            if event.message:
                messagelog = localtimestring(event.timestamp) + "  " + event.message
                messages_pane  = startsc.query_one("#sys-messages-pane")
                messages_pane.post_message(messages_pane.ShowLogs(messagelog))
            # As a system message, there is no change to a devicesc, so nothing further to do
            return

        if (event.eventtype == "Delete") and (not self[event.devicename].enable):
            # As the device has enable False, this Delete event is either requesting an entire
            # device delete, or the last vector of this device is deleted. In either
            # case, this entire device should be deleted
            deviceid = app.itemid.get_devicid(event.devicename)
            if not deviceid:
                # This device is not displayed, nothing to do
                return
            # instruct the startsc to remove the device button
            device_pane = startsc.query_one("#device-pane")
            device_pane.post_message(device_pane.DelButton(deviceid))
            if event.message:
                # show this message as a system message, as there is no device
                # so there is nowhere else to show it
                messagelog = localtimestring(event.timestamp) + "  " + event.message
                messages_pane  = startsc.query_one("#sys-messages-pane")
                messages_pane.post_message(messages_pane.ShowLogs(messagelog))
            # remove all id's associated with this device
            app.itemid.clear_device(event.device)
            # if this device is currently being shown, pop the screen
            if app.itemid.devicename == event.devicename:
                self.clientdata['devicesc'] = None
                app.itemid.devicename = None
                app.pop_screen()
            return


        # handle received events affecting devicesc ################################

        devicesc = self.clientdata.get('devicesc')
        if devicesc is None:
            # A device screen is not currently being shown
            return

        if not event.devicename:
            # This is a system event, already handled
            return

        if not app.itemid.devicename:
            # No device currently being shown
            return

        if event.devicename != app.itemid.devicename:
            # this event refers to a device not currently being shown
            return

        devicename = app.itemid.devicename
        # so this device is currently being shown on devicesc, and the event refers to this device

        # device messages - must be a device message, rather than system message since devicename is given
        if event.eventtype == "Message":
            if event.message:
                messagelog = localtimestring(event.timestamp) + "  " + event.message
                log = devicesc.query_one('#dev-messages-pane')
                log.post_message(log.ShowLogs(messagelog))
            return

        if not event.vectorname:
            return

        vectorid = app.itemid.get_id(event.vectorname)

        if (event.eventtype == "Define" or event.eventtype == "DefineBLOB"):
            if vectorid is None:
                # new vector, add the vector to the tab
                vector = self[devicename][event.vectorname]
                grpid = app.itemid.get_group_id(vector.group)               # if grpid None, a new group has to be created
                if grpid:
                    # The group exists
                    grouptabpane = devicesc.query_one(f"#{grpid}")
                    grouptabpane.post_message(grouptabpane.AddVector(vector))
                else:
                    grouppane = devicesc.query_one("#dev-group-pane")
                    grouppane.post_message(grouppane.AddGroup(vector.group))
                return

        if vectorid is None:
            # no define has been received for this vector, it is not known about
            # should a getproperties be sent here?
            return

        if event.eventtype == "Delete":
            # This vector should be deleted
            vector = self[devicename][event.vectorname]
            grouppane = devicesc.query_one("#dev-group-pane")
            grouppane.post_message(grouppane.DelVector(vector, vectorid))
            # the delete event could include a message, which cannot be displayed on the vector
            # widget, since that will be removed, instead show it on the device message log
            if event.message:
                messagelog = localtimestring(event.timestamp) + "  " + event.message
                log = devicesc.query_one("#dev-messages-pane")
                log.post_message(log.ShowLogs(messagelog))
            return

        # so the vector is currently on display and has a vector pane. The received event may be setting new values
        vectorpane = devicesc.query_one(f"#{vectorid}")

        # Display vector state with timestamp
        if hasattr(event, "state"):
            # shows timestamp and state together
            vectorpane.post_message(vectorpane.ShowTimestamp(localtimestring(event.timestamp)))
            vectorpane.post_message(vectorpane.ShowState(event.state))

        # Display vector message
        if hasattr(event, "message"):
            if event.message:
                vectorpane.post_message(vectorpane.ShowVmessage(localtimestring(event.timestamp) + "  " + event.message))

        if event.eventtype == "TimeOut":
            vectorpane.post_message(vectorpane.SubmitButtonmessage("A Timeout Error has occurred"))
            if vectorpane.vstate == "Busy":
                vectorpane.post_message(vectorpane.ShowTimestamp(localtimestring(event.timestamp)))
                vectorpane.post_message(vectorpane.ShowState("Alert"))
            return


        if event.eventtype not in ("Define", "DefineBLOB", "Set", "SetBLOB"):
            return

        # Only those events with member values now handled

        # For every member in the event, display its value

        for membername, membervalue in event.items():
            mpid = app.itemid.get_id(event.vectorname, membername)
            if not mpid:
                # This member not defined
                continue
            memberpane = vectorpane.query_one(f"#{mpid}")
            if event.vector.vectortype == "SwitchVector":
                memberpane.post_message(memberpane.SetValue(membervalue))
            elif event.vector.vectortype == "LightVector":
                memberpane.post_message(memberpane.SetValue(membervalue))
            elif event.vector.vectortype == "TextVector":
                memberpane.post_message(memberpane.SetValue(membervalue))
            elif event.vector.vectortype == "NumberVector":
                # display a formatted number string rather than the received number
                fvalue = event.vector.getformattedvalue(membername)
                memberpane.post_message(memberpane.SetValue(fvalue))
            elif event.vector.vectortype == "BLOBVector":
                # display the received filename rather than the binary blob received
                # the vector.member() method returns the member given its name
                fvalue = event.vector.member(membername).filename
                memberpane.post_message(memberpane.SetValue(fvalue))



from pathlib import Path
from typing import Iterable

from textual.app import ComposeResult
from textual.widgets import Button, DirectoryTree
from textual.screen import ModalScreen
from textual.containers import Container, Horizontal



class FilteredDirectoryTree(DirectoryTree):
    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        return [path for path in paths if not path.name.startswith(".")]


class FilePane(Container):

    DEFAULT_CSS = """

        FilePane {
            align: center middle;
            width: 70%;
            height: 70%;
            border: mediumvioletred;
            background: $panel;
           }

        FilteredDirectoryTree {
            margin: 1;
           }

        Horizontal {
            height: auto;
            align: right middle;
           }

        Button {
            margin-right: 2;
            }
        """

    def compose(self) -> ComposeResult:
        yield FilteredDirectoryTree("/")
        with Horizontal():
            yield Button("Cancel", variant="primary", id="cancel")
            yield Button("Send", variant="primary", id="send", disabled=True)

    def on_mount(self):
        self.border_title = "Choose File"





class ChooseFileSc(ModalScreen):
    """The class defining the choose file screen."""

    DEFAULT_CSS = """

        ChooseFileSc {
            align: center middle;
            }
        """

    def __init__(self):
        self.selected_filepath = None
        super().__init__()

    def compose(self) -> ComposeResult:
        yield FilePane()

    def on_directory_tree_directory_selected(self, event):
        "On a directory being selected, disable the send button"
        sndbtn = self.query_one("#send")
        sndbtn.disabled = True
        self.selected_filepath = None

    def on_directory_tree_file_selected(self, event):
        """On a file being selected, enable the send button
           and store the filepath"""
        sndbtn = self.query_one("#send")
        sndbtn.disabled = False
        self.selected_filepath = event.path

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "cancel":
            self.dismiss(None)
        if event.button.id == "send":
            self.dismiss(self.selected_filepath)





from textual.widgets import Static, Button, Input, Switch
from textual.reactive import reactive
from textual.containers import Container
from textual.message import Message
from textual.widget import Widget
from textual.css.query import NoMatches

from decimal import Decimal

from indipyclient import getfloat

from .filechooser import ChooseFileSc



class SwitchLabel(Static):

    DEFAULT_CSS = """
        SwitchLabel {
            width: 1fr;
            height: 3;
            content-align: center middle;
        }
        """


class SwitchValue(Static):

    DEFAULT_CSS = """
        SwitchValue {
            width: auto;
            padding: 1;
            height: auto;
            }
        """

    mvalue = reactive("")

    def watch_mvalue(self, mvalue):
        if mvalue:
            if mvalue == "On":
                self.styles.background = "darkgreen"
                self.styles.color = "white"
            elif mvalue == "Off":
                self.styles.background = "red"
                self.styles.color = "white"
            else:
                mvalue = "?"
                self.styles.background = "black"
                self.styles.color = "white"
            self.update(mvalue)



class SwitchMemberPane(Widget):

    DEFAULT_CSS = """
        SwitchMemberPane {
            layout: horizontal;
            background: $panel;
            margin-left: 1;
            margin-bottom: 1;
            height: auto;
        }

        SwitchMemberPane > Container {
            width: 1fr;
            height: auto;
            align: center middle;
        }
        """

    class SetValue(Message):

        def __init__(self, value):
            self.value = value
            super().__init__()

    mvalue = reactive("")

    def __init__(self, vector, member):
        self.member = member
        self.vector = vector
        member_id = self.app.itemid.set_id(vector.name, member.name)
        super().__init__(id=member_id)


    def compose(self):
        "Draw the member"
        if self.member.membervalue:
            self.mvalue = self.member.membervalue

        yield SwitchLabel(self.member.label)
        with Container():
            yield SwitchValue().data_bind(SwitchMemberPane.mvalue)
        with Container():
            if self.member.membervalue == "On":
                if self.vector.perm == "ro":
                    yield Switch(value=True, disabled=True)
                else:
                    yield Switch(value=True)
            else:
                if self.vector.perm == "ro":
                    yield Switch(value=False, disabled=True)
                else:
                    yield Switch(value=False)

    def watch_mvalue(self, mvalue):
        "Alter switches as values received"
        # The displayed SwitchValue will automatically follow mvalue
        # so does not have to be set here.
        # However the switch objects are set here as values are received
        # from iclient
        if self.vector.perm != "ro":
            return
        # Only bother changing switch states if ro
        if not  mvalue:
            return
        try:
            switch = self.query_one(Switch)
        except NoMatches:
            # presumably this vector has not been displayed yet
            return
        if mvalue == "On":
            switch.value = True
        else:
            switch.value = False

    def on_switch_member_pane_set_value(self, message: SetValue) -> None:
        self.mvalue = message.value


class TextLabel(Static):

    DEFAULT_CSS = """
        TextLabel {
            width: 1fr;
            height: 3;
            content-align: center middle;
        }
        """

class ROTextLabel(Static):

    DEFAULT_CSS = """
        ROTextLabel {
            width: 1fr;
            content-align: center middle;
        }
        """

class TextValue(Static):

    DEFAULT_CSS = """
        TextValue {
            width: 1fr;
        }
        """


class ShowText(Container):

    DEFAULT_CSS = """
        ShowText {
            layout: vertical;
            width: 2fr;
            height: auto;
            }

        TextValue {
            width: 1fr;
            padding: 1;
            }

        .textinput {
            layout: horizontal;
            height: auto;
            }

        Button {
            width: auto;
            height: auto;
            }

        """


    def __init__(self, member):
        self.member = member
        super().__init__()

    def compose(self):
        # permission is wo or rw, so show value with editing capbility
        yield TextValue(self.member.membervalue)
        with Container(classes="textinput"):
            yield TextInputField(self.member)
            yield Button("Clear")



class TextMemberPane(Widget):

    DEFAULT_CSS = """
        TextMemberPane {
            layout: horizontal;
            background: $panel;
            margin-left: 1;
            margin-bottom: 1;
            height: auto;
            }
        """

    class SetValue(Message):

        def __init__(self, value):
            self.value = value
            super().__init__()

    def __init__(self, vector, member):
        self.member = member
        self.vector = vector
        member_id = self.app.itemid.set_id(vector.name, member.name)
        super().__init__(id=member_id)


    def compose(self):
        "Draw the member"
        if self.vector.perm == "ro":
            yield ROTextLabel(self.member.label)
            yield TextValue(self.member.membervalue)
            return
        yield TextLabel(self.member.label)
        yield ShowText(self.member)


    def on_text_member_pane_set_value(self, message: SetValue) -> None:
        showtextvalue = self.query_one(TextValue)
        showtextvalue.update(message.value)

    def on_button_pressed(self, event):
        "Clear text input field"
        infld = self.query_one(TextInputField)
        infld.placeholder="Input new text"
        infld.clear()
        event.stop()


class TextInputField(Input):

    DEFAULT_CSS = """
        TextInputField {
            width: 1fr;
            }
        """

    def __init__(self, member):
        self.member = member
        super().__init__(placeholder="Input new text")

    def on_blur(self, event):
        # self.value is the new value input
        if self.value.isprintable():
            checkedvalue = self.value
        else:
            checkedvalue = "Invalid string"
        self.clear()
        self.insert_text_at_cursor(checkedvalue)

    def on_key(self, event):
        if event.character is None:
            return
        # a printable key is pressed
        self.placeholder=""

    def action_submit(self):
        self.screen.focus_next('*')


class LightLabel(Static):

    DEFAULT_CSS = """
        LightLabel {
            width: 1fr;
            height: 3;
            content-align: center middle;
        }
        """

class LightValue(Static):

    DEFAULT_CSS = """
        LightValue {
            padding: 1;
            width: auto;
            height: auto;
        }
        """

    mvalue = reactive("")


    def watch_mvalue(self, mvalue):
        if mvalue:
            if mvalue == "Ok":
                self.styles.background = "darkgreen"
                self.styles.color = "white"
            elif mvalue == "Alert":
                self.styles.background = "red"
                self.styles.color = "white"
            elif mvalue == "Busy":
                self.styles.background = "yellow"
                self.styles.color = "black"
            elif mvalue == "Idle":
                self.styles.background = "black"
                self.styles.color = "white"
            else:
                mvalue = "?"
                self.styles.background = "black"
                self.styles.color = "white"
            self.update(mvalue)


class LightMemberPane(Widget):

    DEFAULT_CSS = """
        LightMemberPane {
            layout: horizontal;
            background: $panel;
            margin-left: 1;
            margin-bottom: 1;
            height: auto;
        }

        LightMemberPane > Container {
            width: 1fr;
            height: auto;
            align: center middle;
            }
        """

    class SetValue(Message):

        def __init__(self, value):
            self.value = value
            super().__init__()

    mvalue = reactive("")

    def __init__(self, vector, member):
        self.member = member
        self.vector = vector
        member_id = self.app.itemid.set_id(vector.name, member.name)
        super().__init__(id=member_id)


    def compose(self):
        "Draw the member"

        if self.member.membervalue:
            self.mvalue = self.member.membervalue

        yield LightLabel(self.member.label)
        with Container():
            yield LightValue().data_bind(LightMemberPane.mvalue)

    def on_light_member_pane_set_value(self, message: SetValue) -> None:
        self.mvalue = message.value


class NumberLabel(Static):

    DEFAULT_CSS = """
        NumberLabel {
            width: 1fr;
            height: 3;
            content-align: center middle;
        }
        """

class NumberValue(Static):

    DEFAULT_CSS = """
        NumberValue {
            width: 1fr;
            height: 3;
            content-align: center middle;
        }
        """

    mvalue = reactive("")

    def watch_mvalue(self, mvalue):
        if mvalue:
            self.update(mvalue)


class NumberMemberPane(Widget):

    DEFAULT_CSS = """
        NumberMemberPane {
            layout: horizontal;
            background: $panel;
            margin-left: 1;
            margin-bottom: 1;
            height: auto;
            }

        NumberMemberPane > Container {
            layout: horizontal;
            width: 2fr;
            height: auto;
            align: center middle;
            }
        """

    class SetValue(Message):

        def __init__(self, value):
            self.value = value
            super().__init__()

    mvalue = reactive("")

    def __init__(self, vector, member):
        self.member = member
        self.vector = vector
        member_id = self.app.itemid.set_id(vector.name, member.name)
        super().__init__(id=member_id)


    def compose(self):
        "Draw the member"
        self.mvalue = self.member.getformattedvalue()
        yield NumberLabel(self.member.label)
        yield NumberValue().data_bind(NumberMemberPane.mvalue)
        if self.vector.perm != "ro":
            with Container():
                yield NumberInputField(self.member, placeholder="Input new number")
                yield Button("Clear")

    def on_button_pressed(self, event):
        "Clear number input field"
        infld = self.query_one(NumberInputField)
        infld.clear()
        event.stop()

    def on_number_member_pane_set_value(self, message: SetValue) -> None:
        self.mvalue = message.value



class NumberInputField(Input):

    DEFAULT_CSS = """

        NumberInputField {
            width: 1fr;
            }
        """

    def __init__(self, member, placeholder):
        self.member = member
        super().__init__(placeholder=placeholder)

    def on_blur(self, event):
        # self.value is the new value input
        if not self.value:
            return
        try:
            newfloat = getfloat(self.value)
        except (ValueError, TypeError):
            self.clear()
            checkedvalue = self.member.getformattedvalue()
            self.insert_text_at_cursor(checkedvalue)
            return
        # check step, and round newfloat to nearest step value
        stepvalue = getfloat(self.member.step)
        minvalue = getfloat(self.member.min)
        if stepvalue:
            stepvalue = Decimal(str(stepvalue))
            difference = newfloat - minvalue
            newfloat = minvalue + float(int(Decimal(str(difference)) / stepvalue) * stepvalue)
        # check not less than minimum
        if newfloat < minvalue:
            # reset input to be the minimum, and accept this
            self.clear()
            checkedvalue = self.member.getformattedstring(minvalue)
            self.insert_text_at_cursor(checkedvalue)
            return
        if self.member.max != self.member.min:
            maxvalue = getfloat(self.member.max)
            if newfloat > maxvalue:
                # reset input to be the maximum, and accept this
                self.clear()
                checkedvalue = self.member.getformattedstring(maxvalue)
                self.insert_text_at_cursor(checkedvalue)
                return
        # reset input to the correct format, and accept this
        self.clear()
        checkedvalue = self.member.getformattedstring(newfloat)
        self.insert_text_at_cursor(checkedvalue)


    def action_submit(self):
        self.screen.focus_next('*')



class BLOBLabel(Static):

    DEFAULT_CSS = """
        BLOBLabel {
            width: 1fr;
            content-align: center middle;
        }
        """

class BLOBRxValue(Static):

    DEFAULT_CSS = """
        BLOBRxValue {
            width: 2fr;
        }
        """

    mvalue = reactive("")

    def watch_mvalue(self, mvalue):
        if mvalue:
            self.update(f"RX data: {mvalue}")


class BLOBTxValue(Static):

    DEFAULT_CSS = """
        BLOBTxValue {
            width: 2fr;
        }
        """


class BlobMemberPane(Widget):

    DEFAULT_CSS = """
        BlobMemberPane {
            layout: horizontal;
            background: $panel;
            margin-left: 1;
            margin-bottom: 1;
            height: auto;
            }

        BlobMemberPane > Container {
            layout: vertical;
            background: $panel;
            width: 2fr;
            height: auto;
            }

        Button {
            margin: 1;
            width: auto;
            height: auto;
            }
        """

    class SetValue(Message):

        def __init__(self, value):
            self.value = value
            super().__init__()

    mvalue = reactive("")

    def __init__(self, vector, member):
        self.member = member
        self.vector = vector
        member_id = self.app.itemid.set_id(vector.name, member.name)
        super().__init__(id=member_id)


    def compose(self):
        "Draw the member"
        yield BLOBLabel(self.member.label)
        # The last filename sent is stored in the member user string
        last_filename = self.member.user_string
        with Container():
            if self.vector.perm == "wo":
                yield BLOBRxValue("RX data: N/A -- Write only --").data_bind(BlobMemberPane.mvalue)
            elif not self.app.blobfolder:
                yield BLOBRxValue("RX data: -- BLOB Folder not set --").data_bind(BlobMemberPane.mvalue)
            elif not self.member.filename:
                yield BLOBRxValue("RX data: -- Nothing yet received --").data_bind(BlobMemberPane.mvalue)
            else:
                yield BLOBRxValue(f"RX data: {self.member.filename}").data_bind(BlobMemberPane.mvalue)
            if self.vector.perm == "ro":
                yield BLOBTxValue("TX data: N/A -- Read only --")
            elif last_filename:
                yield BLOBTxValue(f"TX data: {last_filename}")
                yield Button("Send File")
            else:
                yield BLOBTxValue("TX data: -- No file sent --")
                yield Button("Send File")


    def on_blob_member_pane_set_value(self, message: SetValue) -> None:
        self.mvalue = message.value


    async def on_button_pressed(self, event):
        "Open file chooser screen"
        async def send_path(path):
            if path is not None:
                # memberdict of {membername:(value, blobsize, blobformat)}
                await self.vector.send_newBLOBVector(members={self.member.name:(path, 0, "")})
                self.member.user_string = path.name
                path_text = self.query_one(BLOBTxValue)
                path_text.update(f"TX data: {path.name}")
                # set state to busy
                self.parent.parent.vstate = "Busy"
        self.app.push_screen(ChooseFileSc(), send_path)
        event.stop()
